<?php
namespace Magenest\Salesforce\Model\Sync;

use Magenest\Salesforce\Helper\Data as Helper;
use Magenest\Salesforce\Logger\Logger;
use Magenest\Salesforce\Model\Connector;
use Magenest\Salesforce\Model\Data;
use Magenest\Salesforce\Model\ReportFactory as ReportFactory;
use Magenest\Salesforce\Model\RequestLogFactory;
use Magenest\Salesforce\Model\ResourceModel\Queue as QueueResource;
use Magenest\Salesforce\Model\ResourceModel\Queue\CollectionFactory as QueueCollectionFactory;
use Magenest\Salesforce\Model\ResourceModel\Report\CollectionFactory as ReportCollectionFactory;
use Magento\Config\Model\ResourceModel\Config as ResourceModelConfig;
use Magento\Customer\Model\Customer;
use Magento\Customer\Model\CustomerFactory;
use Magento\Framework\App\Config\ScopeConfigInterface as ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\HTTP\ClientFactory;
use Magento\Framework\Model\ResourceModel\Iterator;
use Magento\Framework\Serialize\Serializer\Json;
use Magento\Framework\Serialize\Serializer\Serialize;

class Contact extends Connector
{
    const SALESFORCE_CONTACT_ATTRIBUTE = 'salesforce_contact_id';

    /** @var CustomerFactory  */
    protected $_customerFactory;

    /** @var Job  */
    protected $_job;

    /** @var DataGetter  */
    protected $dataGetter;

    /** @var Helper  */
    protected $helper;

    /** @var Data  */
    protected $_data;

    /** @var Lead  */
    protected $lead;

    /** @var Account */
    protected $account;

    protected $existedContacts = null;

    protected $createContactIds = null;

    protected $updateContactIds = null;

    /**
     * Contact constructor.
     * @param CustomerFactory $customerFactory
     * @param Data $data
     * @param Job $job
     * @param DataGetter $dataGetter
     * @param Helper $helper
     * @param Lead $lead
     * @param Account $account
     * @param ReportCollectionFactory $reportCollectionFactory
     * @param Helper $helperData
     * @param Json $json
     * @param Serialize $serialize
     * @param ScopeConfigInterface $scopeConfig
     * @param ResourceModelConfig $resourceConfig
     * @param ReportFactory $reportFactory
     * @param QueueCollectionFactory $queueCollectionFactory
     * @param QueueResource $queueResource
     * @param RequestLogFactory $requestLogFactory
     * @param ClientFactory $client
     * @param Iterator $iterator
     * @param Logger $logger
     */
    public function __construct(
        CustomerFactory $customerFactory,
        Data $data,
        Job $job,
        DataGetter $dataGetter,
        Helper $helper,
        Lead $lead,
        Account $account,
        ReportCollectionFactory $reportCollectionFactory,
        Helper $helperData,
        Json $json,
        Serialize $serialize,
        ScopeConfigInterface $scopeConfig,
        ResourceModelConfig $resourceConfig,
        ReportFactory $reportFactory,
        QueueCollectionFactory $queueCollectionFactory,
        QueueResource $queueResource,
        RequestLogFactory $requestLogFactory,
        ClientFactory $client,
        Iterator $iterator,
        Logger $logger
    ) {
        $this->_customerFactory = $customerFactory;
        $this->_data            = $data;
        $this->_type            = 'Contact';
        $this->_table           = 'customer';
        $this->_job             = $job;
        $this->dataGetter       = $dataGetter;
        $this->helper           = $helper;
        $this->lead             = $lead;
        $this->account          = $account;
        parent::__construct(
            $reportCollectionFactory,
            $helperData,
            $json,
            $serialize,
            $scopeConfig,
            $resourceConfig,
            $reportFactory,
            $queueCollectionFactory,
            $queueResource,
            $requestLogFactory,
            $iterator,
            $client,
            $logger
        );
    }

    /**
     * Update or create new a record
     *
     * @param int $id
     * @param boolean $update
     * @return string
     * @throws \Exception
     */
    public function sync($id, $update = false)
    {
        $model = $this->_customerFactory->create()->load($id);
        $customerId = $id;
        $salesforceId = false;
        if ($this->getCustomerOrderCollection($model)->getSize() > 0) {
            $email        = $model->getEmail();
            $salesforceId = $this->searchRecords($this->_type, 'Email', $email);
            if (!$salesforceId && $this->checkSalesforceCustomerStatus($model, "Lead", "Email")) {
                $errorMessage = "This Contact cannot be created, because a Lead with the same email address already existed.";
                $this->saveReport(null, "insert", $this->_type, 2, $errorMessage, $customerId);
                throw new LocalizedException(__($errorMessage));
            }
            if (!$salesforceId || $update) {
                $params = $this->_data->getCustomer($model, $this->_type);
                $params += [
                    'FirstName' => $model->getFirstname(),
                    'LastName'  => $model->getLastname()
                ];
                if ($this->helper->isLinkedContactWithAccount()) {
                    $accountId = $model->getData(Account::SALESFORCE_ACCOUNT_ATTRIBUTE_CODE);
                    if ($accountId == '') {
                        $accountId = $this->account->sync($customerId);
                    }
                    $params = array_replace($params, ['AccountId' => $accountId]);
                }
                $params = array_replace($params, [
                    'Email' => $email
                ]);
                $this->_helperData->saveLog("Contact params mapping: " . $this->_json->serialize($params));
                if ($update && $salesforceId) {
                    $this->updateRecords($this->_type, $salesforceId, $params, $model->getId());
                } else {
                    $salesforceId = $this->createRecords($this->_type, $params, $model->getId());
                }
            }
        } else {
            $this->_helperData->saveLog("Customer has not order");
        }
        return $salesforceId;
    }

    /**
     * @param string $customerEmail
     * @param null $customerId
     * @throws \Exception
     */
    private function _beforeSync($customerEmail, $customerId = null)
    {
        if (!$customerEmail && $customerId) {
            $customerEmail = $this->_customerFactory->create()->load($customerId)->getEmail();
        }
        if ($customerEmail) {
            $this->lead->delete($customerEmail);
        }
    }

    /**
     * Sync by Email
     *
     * @param  $data
     * @return string
     * @throws \Exception
     */
    public function syncByEmail($data)
    {
        $id = $this->searchRecords($this->_type, 'Email', $data['Email']);
        if (!$id) {
            $params = $data;
            $id     = $this->createRecords($this->_type, $params);
        }

        return $id;
    }

    /**
     * sync array of emails data
     *
     * @param $dataEmails
     * @param $operation
     * @return array|mixed|string
     * @throws \Exception
     */
    public function syncByEmails($dataEmails, $operation)
    {
        if (empty($dataEmails)) {
            return [];
        }
        $ids                = [];
        $unsyncedEmails     = [];
        foreach ($dataEmails as $email => $value) {
            $sfId = $this->getCustomerByEmail($email, $this->_type);
            if ($sfId) {
                $ids[$email] = $sfId;
            } else {
                $unsyncedEmails[] = $value;
            }
        }

        if (!empty($unsyncedEmails)) {
            $response = $this->_job->sendBatchRequest($operation, $this->_type, $this->_json->serialize($unsyncedEmails));
            $this->saveReports($operation, $this->_type, $response, $unsyncedEmails, $unsyncedEmails);
            if (is_array($response)) {
                $total = count($response);
                for ($i = 0; $i < $total; $i++) {
                    if (isset($response[$i]['success']) && $response[$i]['success']) {
                        $ids[$unsyncedEmails[$i]['Email']] = $response[$i]['id'];
                    }
                }
            }
        }
        return $ids;
    }

    /**
     * Delete Record
     *
     * @param string $email
     * @throws \Exception
     */
    public function delete($email)
    {
        $contactId = $this->searchRecords('Contact', 'Email', $email);
        if ($contactId) {
            $this->deleteRecords('Contact', $contactId);
        }
    }

    /**
     * @return array|mixed|string|null
     */
    public function syncQueue()
    {
        $createResponse = $this->createContacts();
        $updateResponse = $this->updateContacts();
        $response       = $createResponse + $updateResponse;
        $this->unsetCreateQueue();
        $this->unsetUpdateQueue();
        return $response;
    }

    /**
     * Send request to create contacts
     */
    protected function createContacts()
    {
        $response = [];
        if (!is_null($this->createContactIds)) {
            $response = $this->sendContactsRequest($this->createContactIds, 'insert');
        }
        return $response;
    }

    /**
     * Send request to update contacts
     */
    protected function updateContacts()
    {
        $response = [];
        if (!is_null($this->updateContactIds)) {
            $response = $this->sendContactsRequest($this->updateContactIds, 'update');
        }
        return $response;
    }

    /**
     * @param $customerId
     * @throws \Zend_Http_Client_Exception
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function addRecord($customerId)
    {
        $customer = $this->_customerFactory->create()->load($customerId);
        $id = $this->checkSalesforceCustomerStatus($customer, "Contact", "Email");
        if (!$id) {
            if ($this->checkSalesforceCustomerStatus($customer, "Lead", "Email")) {
                $errorMessage = "This Contact cannot be created,\n
                because a Lead with the same email address already existed.";
                $this->saveReport(null, "insert", $this->_type, 2, $errorMessage, $customerId);
            } else {
                $this->addToCreateQueue($customer);
            }
        } else {
            $this->addToUpdateQueue($customer, $id);
        }
    }

    /**
     * @param \Magento\Customer\Model\Customer $customer
     */
    protected function addToCreateQueue($customer)
    {
        $this->createContactIds[] = [
            'mid'  => $customer->getId(),
            'mObj' => $customer
        ];
    }

    /**
     * @param \Magento\Customer\Model\Customer $customer
     * @param string $salesforceId
     */
    protected function addToUpdateQueue($customer, $salesforceId)
    {
        $this->updateContactIds[] = [
            'mid'  => $customer->getId(),
            'mObj' => $customer,
            'sid'  => $salesforceId
        ];
    }

    protected function unsetCreateQueue()
    {
        $this->createContactIds = null;
    }

    protected function unsetUpdateQueue()
    {
        $this->updateContactIds = null;
    }

    /**
     * @param array $contactIds
     * @param string $operation
     * @return mixed|string
     * @throws \Exception
     */
    public function sendContactsRequest($contactIds, $operation)
    {
        $params = [];
        foreach ($contactIds as $contactId) {
            $customer = $contactId['mObj'];
            if ($this->getCustomerOrderCollection($customer)->getSize() > 0) {
                $info = $this->_data->getCustomer($customer, $this->_type);
                $info += [
                    'FirstName' => $customer->getFirstname(),
                    'LastName'  => $customer->getLastname()
                ];
                if ($this->helper->isLinkedContactWithAccount()) {
                    try {
                        $accountId = $customer->getData(Account::SALESFORCE_ACCOUNT_ATTRIBUTE_CODE);
                        if ($accountId == '') {
                            $accountId = $this->account->sync($customer->getId());
                        }
                        $info = array_replace($info, ['AccountId' => $accountId]);
                    } catch (\Throwable $e) {
                        $this->saveReport(null, 'insert', "Account", 2, $e->getMessage(), $contactId);
                        continue;
                    }
                }
                $info = array_replace($info, [
                    'Email' => $customer->getEmail()
                ]);
                if (isset($contactId['sid'])) {
                    $info += ['Id' => $contactId['sid']];
                }
                $params[] = $info;
            }
        }
        if (!empty($params)) {
            $response = $this->_job->sendBatchRequest($operation, $this->_type, $this->_json->serialize($params));
            $this->saveReports($operation, $this->_type, $response, $contactIds, $params);
            return $response;
        }
        return [];
    }

    /**
     * @param Customer $customer
     * @return \Magento\Sales\Model\ResourceModel\Order\Collection
     */
    protected function getCustomerOrderCollection($customer)
    {
        return $this->helper->getCustomerOrderCollection($customer->getId());
    }
}
